import os
import subprocess
import time
from collections import defaultdict

example_args = {
        "client_encryption":"opc.tcp://localhost:4840 client_cert.der client_key.der server_cert.der",
        "access_control_client_encrypt":"opc.tcp://localhost:4840 client_cert.der client_key.der server_cert.der",
        "client_event_filter":"opc.tcp://localhost:4840",
        "client_connect":"-username user1 -password password -cert client_cert.der -key client_key.der opc.tcp://localhost:4840",
        "pubsub_nodeset_rt_publisher":"-i lo",
        "pubsub_nodeset_rt_subscriber":"-i lo",
        "pubsub_TSN_loopback":"-i lo",
        "pubsub_TSN_loopback_single_thread":"-i lo",
        "pubsub_TSN_publisher":"-i lo",
        "pubsub_TSN_publisher_multiple_thread":"-i lo",
        "server_encryption":"server_cert.der server_key.der client_cert.der",
        "server_loglevel":"--loglevel=1",
        "ci_server":"4840 server_cert.der server_key.der client_cert.der",
        "server_file_based_config":"server_json_config.json5"
        }

server_needed_examples = {
        "access_control_client":"ci_server 4840 server_cert.der server_key.der client_cert.der",
        "access_control_client_encrypt":"ci_server 4840 server_cert.der server_key.der client_cert.der",
        "client":"ci_server 4840 server_cert.der server_key.der client_cert.der",
        "client_async":"ci_server 4840 server_cert.der server_key.der client_cert.der",
        "client_connect":"ci_server 4840 server_cert.der server_key.der client_cert.der",
        "client_connect_loop":"ci_server 4840 server_cert.der server_key.der client_cert.der",
        "client_encryption":"server_encryption server_cert.der server_key.der client_cert.der",
        "client_event_filter":"ci_server 4840 server_cert.der server_key.der client_cert.der",
        "client_historical":"tutorial_server_historicaldata",
        "client_method_async":"ci_server 4840 server_cert.der server_key.der client_cert.der",
        "client_subscription_loop":"ci_server 4840 server_cert.der server_key.der client_cert.der",
        "custom_datatype_client":"custom_datatype_server",
        "discovery_client_find_servers":"discovery_server_lds",
        "discovery_server_register":"discovery_server_lds",
        "pubsub_publish_encrypted":"pubsub_subscribe_encrypted",
        "pubsub_publish_encrypted_sks":"server_pubsub_central_sks server_cert.der server_key.der --enableUnencrypted --enableAnonymous",
        "pubsub_subscribe_encrypted":"pubsub_publish_encrypted",
        "pubsub_subscribe_encrypted_sks":"server_pubsub_central_sks server_cert.der server_key.der --enableUnencrypted --enableAnonymous",
        "pubsub_subscribe_standalone_dataset":"tutorial_pubsub_publish",
        "server_pubsub_publish_rt_level":"server_pubsub_subscribe_rt_level",
        "server_pubsub_rt_information_model":"server_pubsub_subscribe_rt_level",
        "server_pubsub_subscribe_custom_monitoring":"server_pubsub_publish_on_demand",
        "server_pubsub_subscribe_rt_level":"server_pubsub_publish_rt_level",
        "tutorial_client_events":"tutorial_server_events",
        "tutorial_client_firststeps":"tutorial_server_firststeps",
        "tutorial_pubsub_connection":"tutorial_pubsub_subscribe",
        "tutorial_pubsub_mqtt_subscrib":"tutorial_pubsub_mqtt_publish",
        "tutorial_pubsub_publish":"tutorial_pubsub_subscribe",
        "tutorial_pubsub_subscribe":"tutorial_pubsub_publish",
        "tutorial_server_reverseconnect":"ci_server 4841 client_cert.der client_key.der server_cert.der"
}

client_needed_examples = {
        "ci_server":"client",
        "server_ctt":"client",
        "server_inheritance":"client",
        "server_instantiation":"client",
        "server_loglevel":"client",
        "server_mainloop":"client",
        "server_nodeset":"client",
        "server_nodeset_loader":"client",
        "server_nodeset_plcopen":"client",
        "server_nodeset_powerlink":"client",
        "server_settimestamp":"client",
        "server_testnodeset":"client",
        "tutorial_server_monitoreditems":"client",
        "tutorial_server_variabletype":"client",
        "tutorial_server_object":"client",
        "tutorial_server_variable":"client",
        "access_control_server":"access_control_client",
        "custom_datatype_server":"custom_datatype_client",
        "discovery_server_lds":"discovery_client_find_servers",
        "server_encryption":"client_encryption opc.tcp://localhost:4840 client_cert.der client_key.der server_cert.der",
        "server_events_random":"tutorial_client_events opc.tcp://localhost:4840",
        "server_pubsub_central_sks":"pubsub_publish_encrypted_sks",
        "tutorial_server_alarms_conditions":"tutorial_client_events opc.tcp://localhost:4840",
        "tutorial_server_datasource":"tutorial_client_events opc.tcp://localhost:4840",
        "tutorial_server_firststeps":"tutorial_client_firststeps",
        "tutorial_server_historicaldata":"client_historical",
        "tutorial_server_historicaldata_circular":"client_historical",
        "tutorial_server_method":"client_method_async",
        "tutorial_server_method_async":"client_method_async"
}

blacklist = {
        "pubsub_TSN_loopback":1,
        "pubsub_TSN_loopback_single_thread":1,
        "pubsub_TSN_publisher":1,
        "pubsub_TSN_publisher_multiple_thread":1,
        "pubsub_nodeset_rt_publisher":1,
        "pubsub_nodeset_rt_subscriber":1,
        "access_control_client_encrypt":1,
        "client_encryption":1,
        "client_historical":1,
        "client_subscription_loop":1,
        "client_method_async":1,
        "pubsub_subscribe_encrypted":1
}

# Run each example with valgrind.
# Wait 10 seconds and send the SIGINT signal.
# Wait for the process to terminate and collect the exit status.
# Abort when the exit status is non-null.

os.chdir("../build")
current_dir = os.getcwd()
print(f"Current directory: {current_dir}")
example_dir = os.path.join(current_dir, "bin", "examples")
examples = os.listdir(example_dir)

# skipping examples that are in the blacklist
for example in examples:
    if example in blacklist:
        print(f"Skipping {example}")
        continue
    
    # get the arguments for the example
    args = example_args.get(example)
    cmd = ["valgrind", "--errors-for-leak-kinds=all", "--leak-check=full", "--error-exitcode=1", "./bin/examples/"+example]
    if args:
        args_list = args.split()
        cmd += args_list
    print(f"Args for {example}: {cmd} ")

    # check if the example needs a server or client to be running
    server_needed = server_needed_examples.get(example)
    client_needed = client_needed_examples.get(example)
    
    #start the server if needed
    server_process = None
    if server_needed:
        server_cmd = server_needed.split()
        server_cmd[0] = "./bin/examples/" + server_cmd[0]
        print(f"Starting server {server_cmd} from {os.getcwd()}")
        server_process = subprocess.Popen(server_cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        time.sleep(10)
    
    # start the example
    process = subprocess.Popen(cmd)
    time.sleep(10)
    
    # start the client if needed
    client_process = None
    if client_needed:
        client_cmd = client_needed.split()
        client_cmd[0] = "./bin/examples/" + client_cmd[0]
        print(f"Starting client {client_cmd}")
        client_process = subprocess.Popen(client_cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    
    # wait 20 seconds and send the SIGINT signal
    time.sleep(20)
    print(f"Sending SIGINT to {process.pid}")
    process.send_signal(subprocess.signal.SIGINT)
    
    tries = 0
    while process.poll() is None:
        time.sleep(20)
        process.send_signal(subprocess.signal.SIGINT)
        if tries > 5:
            print(f"Process {process.pid} is still running after 5 tries. Terminating process.")
            process.kill()
        tries += 1

    # save the exit code
    exit_code = process.wait()
    if exit_code != 0:
        print(f"Processing {example} failed with exit code {exit_code}")
        exit(exit_code)

    # terminate the server and client
    if server_process:
        print(f"Sending SIGKILL to server {server_process.pid}")
        server_process.kill()
        server_process.wait()
    
    if client_process:
        print(f"Sending SIGKILL to client {client_process.pid}")
        client_process.kill()
        client_process.wait()



#!/usr/bin/env python3

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this 
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import sys
import re

# Converts a header file to restructured text documentation
#
# All text in /** */ comments becomes restructured text. Everything else is
# included as a code-block with C syntax highlighting.
#
# The beginning and end of the header are removed.
# - Find the first /** */ comment -> start of the documentation
# - Find the last line beginning with "#ifdef" -> end of the documentation

remove_keyword = [" UA_EXPORT", " UA_FUNC_ATTR_WARN_UNUSED_RESULT",
                  " UA_FUNC_ATTR_MALLOC", " UA_RESTRICT "]

def clean_comment(line):
    m = re.search("^\s*(\* |/\*\* )(.*?)( \*/)?$", line)
    if not m:
        return "\n"
    return m.group(2) + "\n"

def clean_line(line):
    for keyword in remove_keyword:
        line = line.replace(keyword, "")
    return line

def comment_start(line):
    m = re.search("^\s*/\*\*[ \n]", line)
    if not m:
        return False
    return True

def comment_end(line):
    m = re.search(" \*/$", line)
    if not m:
        return False
    return True

def first_line(c):
    "Searches for the first comment"
    for i in range(len(c)):
        if comment_start(c[i]):
            return i
    return -1

def last_line(c):
    "Searches for the latest ifdef (closing the include guard)"
    reg = re.compile("^#ifdef")
    for i in range(len(c)-1,1,-1):
        if "stop-doc-generation" in c[i]:
            return i
        if "_UA_END_DECLS" in c[i]:
            reg = re.compile("^_UA_END_DECLS")
    last = 1
    for i in range(len(c)-1,1,-1):
        m = reg.match(c[i])
        if m:
            last = i
            break
    # skip empty lines at the end
    for i in range(last-1,1,-1):
        if len(c[i].strip()) > 0:
            return i
    return len(c)-1

if len(sys.argv) < 2:
    print("Usage: python c2rst.py input.c/h output.rst")
    exit(0)

with open(sys.argv[1]) as f:
    c = f.readlines()

with open(sys.argv[2], 'w') as rst:
    in_doc = False
    last = last_line(c)
    for i in range(first_line(c), last+1):
        line = c[i]
        doc_start = False
        doc_end = False
        if in_doc:
            doc_end = comment_end(line)
            line = clean_comment(line)
        else:
            doc_start = comment_start(line)
            if doc_start:
                doc_end = comment_end(line)
                line = clean_comment(line)

        if doc_start:
            in_doc = True

        if not ((doc_start or doc_end) and line == "\n"):
            if not in_doc:
                line = "   " + line
            rst.write(clean_line(line))

        if doc_end and i < last:
            rst.write("\n.. code-block:: c\n\n")
            in_doc = False



#!/usr/bin/env python3

### This Source Code Form is subject to the terms of the Mozilla Public
### License, v. 2.0. If a copy of the MPL was not distributed with this
### file, You can obtain one at http://mozilla.org/MPL/2.0/.

###    Copyright 2023 (c) Fraunhofer IOSB (Author: Noel Graf)

import argparse
import re
import os
import platform

parser = argparse.ArgumentParser()
parser.add_argument('outfile', help='outfile w/o extension')
parser.add_argument('namespace', help='namespace')
args = parser.parse_args()

include_line_counter = 0
code_line_counter = 0
existing_namespaces = []
data = []


def is_unix():
    try:
        # Check if the os.uname() function is available
        os.uname()
        return True
    except:
        return False


def windows_exec():
    import msvcrt

    global include_line_counter
    global code_line_counter
    global existing_namespaces
    global data

    # Open the file in read-write mode
    with open(args.outfile + ".c", "r+", encoding='utf8') as file:
        # Lock the file
        msvcrt.locking(file.fileno(), msvcrt.LK_LOCK, 1)

        # Write to the file
        write_code_generation(file)

        # Unlock the file
        msvcrt.locking(file.fileno(), msvcrt.LK_UNLCK, 1)


def unix_exec():
    import fcntl

    global include_line_counter
    global code_line_counter
    global existing_namespaces
    global data

    with open(args.outfile + ".c", "r+", encoding='utf8') as file:
            # Acquire a lock on the file
            fcntl.flock(file, fcntl.LOCK_EX)

            # Write to the file
            write_code_generation(file)

            # Release the lock on the file
            fcntl.flock(file, fcntl.LOCK_UN)


def print_include(string):
    data[include_line_counter] = string


def print_function_call(string):
    data[code_line_counter] = string


def generate_code():
    #########################
    # Print the header file #
    #########################

    print_include(u'''
#include <open62541/{namespace}.h>
//<
'''.format(namespace=args.namespace))

    #########################
    # Print the source file #
    #########################

    print_function_call(u'''
/* {namespace} */
retval |= {namespace}(server);
if(retval != UA_STATUSCODE_GOOD) {{
UA_LOG_ERROR(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Adding the {namespace} failed. Please check previous error output.");
UA_Server_delete(server);
return EXIT_FAILURE;
}}
UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "The {namespace} successfully added.");
//>
'''.format(namespace=args.namespace))


def write_code_generation(file):
    global include_line_counter
    global code_line_counter
    global existing_namespaces
    global data

    # Write to the file
    lines = file.readlines()
    line_counter = 0
    # List which contains the existing namespaces
    existing_namespaces = []
    for line in lines:
        namespaces = re.findall("namespace_.*_generated\(.*\)", line)
        if namespaces:
            namespaces = namespaces[0].split("(")[0]
            existing_namespaces.append(namespaces)
        if re.search("//<", line):
            include_line_counter = line_counter
        if re.search("//>", line):
            code_line_counter = line_counter
        line_counter += 1

    if args.namespace not in existing_namespaces:
        # Set the file descriptor to the beginning of the file
        file.seek(0)
        # read a list of lines into data
        data = file.readlines()
        generate_code()
        # Set the file descriptor to the beginning of the file and delete content
        file.seek(0)
        file.truncate(0)
        file.writelines(data)


if platform.system == "Linux":
    unix_exec()
else:
    if is_unix():
        unix_exec()
    else:
        windows_exec()



#!/usr/bin/env python3

### This Source Code Form is subject to the terms of the Mozilla Public
### License, v. 2.0. If a copy of the MPL was not distributed with this
### file, You can obtain one at http://mozilla.org/MPL/2.0/.

###    Copyright 2023 (c) Fraunhofer IOSB (Author: Noel Graf)

import argparse
from io import open

parser = argparse.ArgumentParser()
parser.add_argument('outfile', help='outfile w/o extension')
args = parser.parse_args()

fh = open(args.outfile + ".h", "wt", encoding='utf8')
fc = open(args.outfile + ".c", "wt", encoding='utf8')

def printh(string):
    print(string, end=u'\n', file=fh)
def printc(string):
    print(string, end=u'\n', file=fc)

#########################
# Print the header file #
#########################

printh(u'''
/* WARNING: This is a generated file.
 * Any manual changes will be overwritten. */
 
#ifndef NODESETINJECTOR_H_
#define NODESETINJECTOR_H_

#include <open62541/server.h>
#include <open62541/plugin/log_stdout.h>

_UA_BEGIN_DECLS

extern UA_StatusCode UA_Server_injectNodesets(UA_Server *server);

_UA_END_DECLS

#endif /* NODESETINJECTOR_H_ */
''')

#########################
# Print the source file #
#########################

printc(u'''
/* WARNING: This is a generated file.
 * Any manual changes will be overwritten. */

#include "nodesetinjector.h"
//<

UA_StatusCode UA_Server_injectNodesets(UA_Server *server) {
UA_StatusCode retval = UA_STATUSCODE_GOOD;
UA_LOG_INFO(UA_Log_Stdout, UA_LOGCATEGORY_SERVER, "Attaching the AUTOLOAD Nodesets to the server!");
//>

return retval;
}
''')

fc.close()
fh.close()



#!/usr/bin/env python3

# Load into gdb with 'source <path-to>/tools/gdb-prettyprint.py'
# Make sure to have 'set print pretty on' to get nice structure printouts

import base64
import gdb

def findType(name):
    tt = None
    try:
        tt = gdb.lookup_type("UA_" + name)
    except Exception:
        try:
            tt = gdb.lookup_type(name)
        except Exception:
            pass
    return tt

class String:
    def __init__(self, val):
        self.val = val

    def to_string(self):
        data = self.val['data']
        if int(data) == 0:
            return "UA_STRING_NULL"
        length = int(self.val['length'])
        inferior = gdb.selected_inferior()
        return "\"%s\"" % inferior.read_memory(data, length).tobytes().decode(errors='replace')

class ByteString:
    def __init__(self, val):
        self.val = val

    @staticmethod
    def print(s):
        length = int(s['length'])
        data = s['data']
        if int(data) == 0:
            return "UA_BYTESTRING_NULL"
        inferior = gdb.selected_inferior()
        encoded = base64.b64encode(inferior.read_memory(data, length).tobytes())
        return "\"%s\"" % encoded.decode(errors='replace')

    def to_string(self):
        data = self.val['data']
        if int(data) == 0:
            return "UA_BYTESTRING_NULL"
        return "UA_ByteString(%s)" % ByteString.print(self.val)

class LocalizedText:
    def __init__(self, val):
        self.val = val

    def to_string(self):
        return "UA_LocalizedText(%s, %s)" % (self.val['locale'], self.val['text'])

class QualifiedName:
    def __init__(self, val):
        self.val = val

    def to_string(self):
        return "UA_QualifiedName(%i, %s)" % (self.val['namespaceIndex'], self.val['name'])

class Guid:
    def __init__(self, val):
        self.val = val

    @staticmethod
    def print(g):
        data = (g['data1'], g['data2'], g['data3'], g['data4'][0],
                g['data4'][1], g['data4'][2], g['data4'][3], g['data4'][4],
                g['data4'][5], g['data4'][6], g['data4'][7])
        return "%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x" % data

    def to_string(self):
        return "UA_Guid(%s)" % Guid.print(self.val)

class NodeId:
    def __init__(self, val):
        self.val = val

    @staticmethod
    def print(val):
        s = ""
        ns = int(val['namespaceIndex'])
        idType = int(val['identifierType'])
        if ns > 0:
            s += "ns=%i;" % ns
        if idType < 3:
            s += "i=%i" % val['identifier']['numeric']
        elif idType == 3:
            s += "s=%s" % val['identifier']['string']
        elif idType == 4:
            s += "g=%s" % Guid.print(val['identifier']['guid'])
        elif idType == 5:
            s += "b=%s" % ByteString.print(val['identifier']['byteString'])
        else:
            raise gdb.GdbError("Invalid NodeId Identifier Type")
        return s

    def to_string(self):
        return "UA_NodeId(%s)" % NodeId.print(self.val)

class ExpandedNodeId:
    def __init__(self, val):
        self.val = val

    def to_string(self):
        s = ""
        if self.val['serverIndex'] > 0:
            s += "svr=%i;" % self.val['serverIndex']
        if self.val['namespaceUri']['data'] > 0:
            s += "nsu=%s;" % self.val['namespaceUri']
        s += NodeId.print(self.val['nodeId'])
        return "UA_ExpandedNodeId(%s)" % s

class ExtensionObject:
    def __init__(self, val):
        self.val = val

    def to_string(self):
        encoding = self.val['encoding']
        content = self.val['content']
        if encoding == 0:
            return "UA_ExtensionObject()"
        elif encoding == 1 or encoding == 2:
            encoded = content['encoded']
            return "UA_ExtensionObject(%s, %s)" % (encoded['typeId'], encoded['body'])
        decoded = content['decoded']
        if int(decoded['type']) == 0:
            return "UA_ExtensionObject()"
        datatype = decoded['type'].dereference()
        typeName = datatype['typeName'].string()
        tt = findType(typeName)
        if not tt:
            return "UA_ExtensionObject<%s>((void*) %s)" % (typeName, decoded['data'])
        value = decoded['data'].cast(tt.pointer()).dereference()
        return "UA_ExtensionObject<%s>(%s)" % (tt, value)

class Variant:
    def __init__(self, val):
        self.val = val

    def to_string(self):
        if int(self.val['type']) == 0:
            return "UA_Variant()"
        datatype = self.val['type'].dereference()
        tt = findType(datatype['typeName'].string())
        if not tt:
            return "UA_Variant()"
        data_ptr = int(self.val['data'])
        array_length = int(self.val['arrayLength'])
        content = None
        if array_length == 0 and data_ptr > 1:
            content = self.val['data'].cast(tt.pointer()).dereference()
        else:
            content = self.val['data'].cast(tt.array(array_length - 1).pointer()).dereference()
        dims_length = int(self.val['arrayDimensionsSize'])
        if dims_length == 0:
            return "UA_Variant<%s>(%s)" % (tt, content)
        dims_type = findType("UInt32").array(dims_length - 1).pointer()
        dims = self.val['arrayDimensions'].cast(dims_type).dereference()
        return "UA_Variant<%s[%i]>(%s, arrayDimensions = %s)" % (tt, array_length, content, dims)

def build_open62541_printer():
   pp = gdb.printing.RegexpCollectionPrettyPrinter("open62541")
   pp.add_printer('UA_String', '^UA_String$', String)
   pp.add_printer('UA_ByteString', '^UA_ByteString$', ByteString)
   pp.add_printer('UA_LocalizedText', '^UA_LocalizedText$', LocalizedText)
   pp.add_printer('UA_QualifiedName', '^UA_QualifiedName$', QualifiedName)
   pp.add_printer('UA_Guid', '^UA_Guid$', Guid)
   pp.add_printer('UA_NodeId', '^UA_NodeId$', NodeId)
   pp.add_printer('UA_ExpandedNodeId', '^UA_ExpandedNodeId$', ExpandedNodeId)
   pp.add_printer('UA_ExtensionObject', '^UA_ExtensionObject$', ExtensionObject)
   pp.add_printer('UA_Variant', '^UA_Variant$', Variant)
   return pp

gdb.printing.register_pretty_printer(gdb.current_objfile(),
                                     build_open62541_printer())



#!/usr/bin/env python3

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import subprocess
import os
import re
from email.utils import formatdate
import datetime
import shutil

# Get the base filesystem paths.
# The current path is used as the "build environment"
dirpath = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)),".."))
debian_path = os.path.join(dirpath, "tools/packaging/debian")
target_debian_path = os.path.join(dirpath, "debian")
shutil.copytree(debian_path, target_debian_path, ignore=shutil.ignore_patterns('*-template'))

# Unpack the library version from the git information
git_describe_version = subprocess.check_output(["git", "describe", "--tags", "--dirty", "--match", "v*"]).decode('utf-8').strip()

# v1.2
# v1.2.3
# v1.2.3-rc1
# v1.2.3-rc1-dirty
# v1.2.3-5-g4538abcd
# v1.2.3-5-g4538abcd-dirty
# git_describe_version = "v1.2.3"

m = re.match(r"^v([0-9]{1,4})(\.[0-9]{1,4})?(\.[0-9]{1,4})?(-(.*){1,100})?$", git_describe_version)
version_major = m.group(1) if m.group(1) is not None else "0"
version_minor = m.group(2).replace(".", "") if m.group(2) is not None else "0"
version_patch = m.group(3).replace(".", "") if m.group(3) is not None else "0"
version_label = m.group(4) if m.group(4) is not None else ""
#print("major {} minor {} patch {} label {}".format(version_major, version_minor, version_patch, version_label))

# We can not use unstable for now, because dpkg-buildpackage wants to sign them
# It will fail with gpg: skipped "open62541 Team <open62541-core@googlegroups.com>": No secret key
#debian_distribution = "unstable"
#if version_label is not "":
debian_distribution = "UNRELEASED"

changelog_file = os.path.join(debian_path, "changelog")

# remove leading 'v'
changelog_version = git_describe_version[1:] if git_describe_version[0] == 'v' else git_describe_version
# replace all '-' with '~' in version
changelog_version = changelog_version.replace('-', '~')

# prefix the version string with the current ISO datetime to ensure correct version ordering.
# See https://github.com/open62541/open62541/issues/3140
changelog_version = datetime.datetime.utcnow().replace(microsecond=0).isoformat().replace('-', '').replace(':', '') + '~' + changelog_version

# Create an updated changelog file with the version information
with open(changelog_file, 'r') as original:
    data = original.read()
with open(os.path.join(target_debian_path, "changelog"), 'w') as modified:
    new_entry = """open62541 ({version}) {distribution}; urgency=medium

  * Full changelog is available here:
    https://github.com/open62541/open62541/blob/master/CHANGELOG

 -- open62541 Team <open62541-core@googlegroups.com>  {time}
""".format(version=changelog_version, time=formatdate(), distribution = debian_distribution)

    modified.write(new_entry + "\n" + data)

# Create control file and replace template variables
control_file_template = os.path.join(debian_path, "control-template")
control_file = os.path.join(target_debian_path, "control")
shutil.copy(control_file_template, control_file)
with open(control_file, 'r+') as f:
    content = f.read()
    f.seek(0)
    f.truncate()
    f.write(content.replace('<soname>', "{}.{}".format(version_major, version_minor)))

# rename the install template to match the soname
install_file_template = os.path.join(debian_path, "libopen62541.install-template")
install_file = os.path.join(target_debian_path, "libopen62541-{}.{}.install".format(version_major, version_minor))
shutil.copy(install_file_template, install_file)

install_file_template = os.path.join(debian_path, "libopen62541-dev.install-template")
install_file = os.path.join(target_debian_path, "libopen62541-{}.{}-dev.install".format(version_major, version_minor))
shutil.copy(install_file_template, install_file)

install_file_template = os.path.join(debian_path, "libopen62541-tools.install-template")
install_file = os.path.join(target_debian_path, "libopen62541-{}.{}-tools.install".format(version_major, version_minor))
shutil.copy(install_file_template, install_file)

# Update CMakeLists.txt to include full version string
with open(os.path.join(dirpath,"CMakeLists.txt"), 'r+') as f:
    lines = f.readlines()
    f.seek(0)
    f.truncate()
    for idx, line in enumerate(lines):
        if idx == 1:
            f.write('set(OPEN62541_VERSION "{}")\n'.format(git_describe_version))
        f.write(line)



#!/usr/bin/env python3

### This Source Code Form is subject to the terms of the Mozilla Public
### License, v. 2.0. If a copy of the MPL was not distributed with this
### file, You can obtain one at http://mozilla.org/MPL/2.0/.

###    Copyright 2023 (c) Fraunhofer IOSB (Author: Noel Graf)

from __future__ import print_function
import sys
import xml.dom.minidom as dom
import codecs
import re
import base64
import argparse

###############################
# Parse the Command Line Input#
###############################

parser = argparse.ArgumentParser()
parser.add_argument('-x', '--xml',
                    metavar="<nodeSetXML>",
                    dest="xmlfile",
                    help='NodeSet XML file with nodes that shall be generated.')

parser.add_argument('outputFile',
                    metavar='<outputFile>',
                    help='output file w/o extension')

args = parser.parse_args()

# Extract the BSD Blob from the XML file.
nodeset_base = open(args.xmlfile, "rb")
fileContent = nodeset_base.read()
# Remove BOM since the dom parser cannot handle it on python 3 windows
if fileContent.startswith(codecs.BOM_UTF8):
    fileContent = fileContent.lstrip(codecs.BOM_UTF8)
if (sys.version_info >= (3, 0)):
    fileContent = fileContent.decode("utf-8")

# Remove the uax namespace from tags. UaModeler adds this namespace to some elements
fileContent = re.sub(r"<([/]?)uax:(.+?)([/]?)>", "<\\g<1>\\g<2>\\g<3>>", fileContent)

nodesets = dom.parseString(fileContent).getElementsByTagName("UANodeSet")
if len(nodesets) == 0 or len(nodesets) > 1:
    raise Exception("contains no or more then 1 nodeset")
nodeset = nodesets[0]
variableNodes = nodeset.getElementsByTagName("UAVariable")
for nd in variableNodes:
    if (nd.hasAttribute("SymbolicName") and (re.match(r".*_BinarySchema", nd.attributes["SymbolicName"].nodeValue) or nd.attributes["SymbolicName"].nodeValue == "TypeDictionary_BinarySchema")) or (nd.hasAttribute("ParentNodeId") and not nd.hasAttribute("SymbolicName") and re.fullmatch(r"i=93", nd.attributes["ParentNodeId"].nodeValue)):
        type_content = nd.getElementsByTagName("Value")[0].getElementsByTagName("ByteString")[0]
        with open(args.outputFile, 'w') as f:
            f.write(base64.b64decode(type_content.firstChild.nodeValue).decode("utf-8"))
    elif nd.hasAttribute("BrowseName") and nd.getAttribute("BrowseName").endswith("TypeDictionary"):
        references = nd.getElementsByTagName("Reference")
        for ref in references:
            if ref.getAttribute("ReferenceType") == "HasComponent" and ref.firstChild.nodeValue == "i=93":
                type_content = nd.getElementsByTagName("Value")[0].getElementsByTagName("ByteString")[0]
                with open(args.outputFile, 'w') as f:
                    f.write(base64.b64decode(type_content.firstChild.nodeValue).decode("utf-8"))
                break



#!/usr/bin/env python3

# coding: UTF-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# This script checks the valgrind output for errors.
# The track-fds does not cause an error if there are too many FDs open,
# therefore we parse the output and fail if there are open FDs

import sys
import subprocess
import os.path
import re
import os

logfile = sys.argv[1]

valgrind_command = ' '.join('"' + item + '"' for item in sys.argv[2:])

# Execute a command and output its stdout text
def execute(command):

    process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    # Poll process for new output until finished
    while True:
        nextline = process.stdout.readline().decode('utf-8')
        if nextline == '' and process.poll() is not None:
            break
        sys.stdout.write(nextline)
        sys.stdout.flush()

    return process.returncode

ret_code = execute(valgrind_command)

if not os.path.isfile(logfile):
    print("### PYTHON ERROR: Valgrind logfile does not exist: " + logfile)
    exit(1)

log_content = ""
with open(logfile, 'r') as content_file:
    log_content = content_file.read()

if len(log_content) == 0:
    print("### PYTHON ERROR: Valgrind logfile is empty: " + logfile)
    exit(1)

# Remove output of possible bug in OSX
# --16672-- run: /usr/bin/dsymutil "/Users/travis/build/Pro/open62541/build/bin/tests/check_utils"
# --16672-- WARNING: unhandled amd64-darwin syscall: unix:464
# --16672-- You may be able to write your own handler.
# --16672-- Read the file README_MISSING_SYSCALL_OR_IOCTL.
# --16672-- Nevertheless we consider this a bug.  Please report
# --16672-- it at http://valgrind.org/support/bug_reports.html.
replace_re = re.compile(r"^--(\d+)--\s+run: .*-- it at http://valgrind.org/support/bug_reports.html\.$\n", re.MULTILINE | re.DOTALL)
log_content = replace_re.sub('', log_content)

# Try to parse the output. Look for the following line:
# ==17054== FILE DESCRIPTORS: 5 open at exit.
descriptors_re = re.compile(r".*==(\d+)==\s+FILE DESCRIPTORS: (\d+) open at exit\..*")
m = descriptors_re.match(log_content)

if not m:
    print("### PYTHON ERROR: File descriptors header not found: " + logfile)
    print(log_content)
    exit(1)

log_content = descriptors_re.sub('', log_content)

valgrind_number = m.group(1)
open_count = int(m.group(2))

# Remove the open file descriptors which are inherited from parent. they look like this:
#==21343== Open file descriptor 3: /home/user/open62541/build/bin/tests/discovery.log
#==21343==    <inherited from parent>
#==21343==
#==21343== Open file descriptor 2:
#==21343==    <inherited from parent>
#==21343==
#==21343== Open file descriptor 1:
#==21343==    <inherited from parent>
#==21343==
#==21343== Open file descriptor 0: /dev/pts/1
#==21343==    <inherited from parent>

replace_re = re.compile(r"^==" + str(valgrind_number) + r"==\s+Open .*$\n" +
                        r"^==" + str(valgrind_number) + r"==\s+<inherited from parent>$\n" +
                        r"(^==" + str(valgrind_number) + r"==\s+$\n)*", re.MULTILINE)
log_content = replace_re.sub('', log_content)

# Valgrind detected a memleak if ret_code != 0
if ret_code != 0:
    print(log_content)
    exit(ret_code)

# No issues by valgrind
if len(log_content) == 0 or log_content.isspace():
    exit(0)

# There is something fishy in the valgrind output, so error-exit
print(log_content)
exit(1)



#!/usr/bin/env python3

# coding: UTF-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this 
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import print_function
import argparse
import os.path
import re
import io

parser = argparse.ArgumentParser()
parser.add_argument('version', help='file version')
parser.add_argument('outfile', help='outfile with extension .c/.h')
parser.add_argument('inputs', nargs='*', action='store', help='input filenames')
args = parser.parse_args()

outname = args.outfile.split("/")[-1]
is_c = False
if outname[-2:] == ".c":
    is_c = True
pos = outname.find(".")
if pos > 0:
    outname = outname[:pos]
include_re = re.compile("^#[\s]*include (\".*\").*$|^#[\s]*include (<open62541/.*>).*$")
guard_re = re.compile("^#(?:(?:ifndef|define)\s*[A-Z_]+_H_|endif /\* [A-Z_]+_H_ \*/|endif // [A-Z_]+_H_|endif\s*/\*\s*!?[A-Z_]+_H[_]+\s*\*/)")

print ("Starting amalgamating file "+ args.outfile)

file = io.open(args.outfile, 'wt', encoding='utf8', errors='replace')
file.write(u"""/* THIS IS A SINGLE-FILE DISTRIBUTION CONCATENATED FROM THE OPEN62541 SOURCES
 * visit http://open62541.org/ for information about this software
 * Git-Revision: %s
 */

/*
 * Copyright (C) 2014-2021 the contributors as stated in the AUTHORS file
 *
 * This file is part of open62541. open62541 is free software: you can
 * redistribute it and/or modify it under the terms of the Mozilla Public
 * License v2.0 as stated in the LICENSE file provided with open62541.
 *
 * open62541 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.
 */\n\n""" % args.version)

if is_c:
    file.write(u'''#ifndef UA_DYNAMIC_LINKING_EXPORT
# define UA_DYNAMIC_LINKING_EXPORT
# define MDNSD_DYNAMIC_LINKING
#endif

#define UA_INLINABLE_IMPL 1

/* Disable security warnings for BSD sockets on MSVC */
#if defined(_MSC_VER) && !defined(_CRT_SECURE_NO_WARNINGS)
# define _CRT_SECURE_NO_WARNINGS
#endif

#include "%s.h"
''' % outname)
else:
    file.write(u'''#ifndef %s
#define %s
''' % (outname.upper() + u"_H_", outname.upper() + u"_H_"))

# Remove the filesystem folder prefix
initial = 999
for fname in args.inputs:
    pos = fname.find("include")
    if pos < 0:
        pos = fname.find("src")
    if pos < 0:
        continue
    if pos - 1 < initial:
        initial = pos - 1

for fname in args.inputs:
    with io.open(fname, encoding='utf8', errors='replace') as infile:
        file.write(u"\n/**** amalgamated original file \"" + fname[initial:] + u"\" ****/\n\n")
        print ("Integrating file '" + fname + "' ... ", end=""),
        for line in infile:
            inc_res = include_re.match(line)
            guard_res = guard_re.match(line)
            if not inc_res and not guard_res:
                file.write(line)
        # Ensure file is written to disk.
        file.flush()
        os.fsync(file.fileno())
        print ("done."),

if not is_c:
    file.write(u"#endif /* %s */\n" % (outname.upper() + u"_H_"))

# Ensure file is written to disk.
# See https://stackoverflow.com/questions/13761961/large-file-not-flushed-to-disk-immediately-after-calling-close
file.flush()
os.fsync(file.fileno())
file.close()

print ("The size of "+args.outfile+" is "+ str(os.path.getsize(args.outfile))+" Bytes.")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import subprocess


def try_get_branch():
    try:
        git_res = subprocess.run(['git', 'branch', '--show-current'], capture_output=True, text=True)
        if git_res.returncode != 0:
            return None
        return git_res.stdout.rstrip()
    except OSError:
        return None


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

version = try_get_branch() or '${OPEN62541_VER_MAJOR}.${OPEN62541_VER_MINOR}.${OPEN62541_VER_PATCH}' or "unknown"
release = '${OPEN62541_VER_MAJOR}.${OPEN62541_VER_MINOR}.${OPEN62541_VER_PATCH}${OPEN62541_VER_LABEL}'

project = 'open62541 (' + version + ')'
copyright = '2025, The open62541 authors'
author = 'The open62541 authors'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.graphviz',
    'sphinx_rtd_theme',
]

templates_path = ['_templates']
exclude_patterns = []

language = 'en'

master_doc = 'toc'

numfig = True

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

html_logo = 'open62541_html.png'

pygments_style = 'sphinx' # TODO makes code blocks ugly (overrides sphinx_rtd_theme)

# -- Options for LaTeX output ---------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-latex-output

latex_logo = 'open62541.png'

latex_documents = [
  # (startdocname,   targetname,      title,                     author, theme,    toctree_only)
    (master_doc,     'open62541.tex', 'open62541 Documentation', author, 'manual', False),
]

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
'papersize': 'a4paper',

# Font settings
'fontpkg': r"""
\usepackage[scaled=.98,varqu,varl]{zi4}
\usepackage[default,semibold]{sourcesanspro}
""",

# The font size ('10pt', '11pt' or '12pt').
'pointsize': '11pt',

# Clean up the header and footer
'preamble': r"""
\fancypagestyle{normal}{
    \fancyhf{}
    \fancyfoot[RO,LE]{\thepage}
    \renewcommand{\headrulewidth}{0pt}
    \renewcommand{\footrulewidth}{0pt}
}
\fancypagestyle{plain}{
    \fancyhf{}
    \fancyfoot[RO,LE]{\thepage}
    \renewcommand{\headrulewidth}{0pt}
    \renewcommand{\footrulewidth}{0pt}
}""",

'maketitle': r"""
    \makeatletter
 \begin{titlepage}%
      \begingroup % for PDF information dictionary
       \def\endgraf{ }\def\and{\& }%
       \pdfstringdefDisableCommands{\def\\{, }}% overwrite hyperref setup
       \hypersetup{pdfauthor={\@author}, pdftitle={\@title}}%
      \endgroup
    \begin{flushright}%
	\vspace*{2\baselineskip}
      \begin{minipage}{.75\linewidth}
      \sphinxlogo
      \end{minipage} \bigskip \par
      \py@HeaderFamily
      {\Huge \@title \par}
      {\itshape\LARGE \py@release\releaseinfo \par}
      \bigskip
      {\large \@date \par}%
    \end{flushright}
  \end{titlepage}%
  \if@openright\cleardoublepage\else\clearpage\fi
\makeatother
""",

# Latex figure (float) alignment
#'figure_align': 'htbp',
}

# -- Options for manual page output ---------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-manual-page-output

man_pages = [
  # (source start file, name,        description,               authors,  manual section)
    (master_doc,        'open62541', 'open62541 Documentation', [author], 1)
]

# -- Options for Texinfo output -------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-texinfo-output

texinfo_documents = [
# (source start file, target name, title,                     author, dir menu entry, description,                        category)
  (master_doc,        'open62541', 'open62541 Documentation', author, 'open62541',    'One line description of project.', 'Miscellaneous'),
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# It is based on the idea of http://0pointer.net/blog/projects/copyright.html

import os
import re
import io
import sys

from git import *
from shutil import move

if sys.version_info[0] >= 3:
    # strings are already parsed to unicode
    def unicode(s):
        return s


# Replace the name by another value, i.e. add affiliation or replace user name by full name
# only use lower case name
authorFullName = {
    'staldert': 'Thomas Stalder, Blue Time Concept SA',
    'mark giraud': 'Mark Giraud, Fraunhofer IOSB',
    'julius pfrommer': 'Julius Pfrommer, Fraunhofer IOSB',
    'stefan profanter': 'Stefan Profanter, fortiss GmbH',
}

# Skip commits with the following authors, since they are not valid names
# and come from an invalid git config
skipNames = ['=', 'open62541', 'opcua']

def compactYears(yearList):

    current = None
    last = None
    result = []

    for yStr in yearList:
        y = int(yStr)
        if last is None:
            current = y
            last = y
            continue

        if y == last + 1:
            last = y
            continue

        if last == current:
            result.append("%i" % last)
        else:
            result.append("%i-%i" % (current, last))

        current = y
        last = y

    if not last is None:
        if last == current:
            result.append("%i" % last)
        else:
            result.append("%i-%i" % (current, last))

    return ", ".join(result)

fileAuthorStats = dict()

def insertCopyrightAuthors(file, authorsList):
    copyrightEntries = list()
    for author in authorsList:
        copyrightEntries.append(unicode("Copyright {} (c) {}").format(compactYears(author['years']), author['author']))

    copyrightAdded = False
    commentPattern = re.compile(r"(.*)\*/$")

    tmpName = file + ".new"
    tempFile = io.open(tmpName, mode="w", encoding="utf-8")
    with io.open(file, mode="r", encoding="utf-8") as f:
        for line in f:
            if copyrightAdded or not commentPattern.match(line):
                tempFile.write(line)
            else:
                tempFile.write(commentPattern.match(line).group(1) + "\n *\n")
                for e in copyrightEntries:
                    tempFile.write(unicode(" *    {}\n").format(e))
                tempFile.write(unicode(" */\n"))
                copyrightAdded = True
    tempFile.close()
    os.unlink(file)
    move(tmpName, file)

def updateCopyright(repo, file):
    print("Checking file {}".format(file))

    # Build the info on how many lines every author commited every year
    relativeFilePath = file[len(repo.working_dir)+1:].replace("\\","/")

    if not relativeFilePath in fileAuthorStats:
        print("File not found in list: {}".format(relativeFilePath))
        return

    stats = fileAuthorStats[relativeFilePath]

    # Now create a sorted list and filter out small contributions
    authorList = list()

    for author in stats:
        if author in skipNames:
            continue

        authorYears = list()
        for year in stats[author]['years']:
            if stats[author]['years'][year] < 10:
                # ignore contributions for this year if less than 10 lines changed
                continue
            authorYears.append(year)
        if len(authorYears) == 0:
            continue
        authorYears.sort()

        if author.lower() in authorFullName:
            authorName = authorFullName[author.lower()]
        else:
            authorName = author


        authorList.append({
            'author': authorName,
            'years': authorYears,
            'first_commit': stats[author]['first_commit']
        })

    # Sort the authors list first by year, and then by name

    authorListSorted = sorted(authorList, key=lambda a: a['first_commit'])
    insertCopyrightAuthors(file, authorListSorted)


# This is required since some commits use different author names for the same person
assumeSameAuthor = {
    'Mark': u'Mark Giraud',
    'Infinity95': u'Mark Giraud',
    'janitza-thbe': u'Thomas Bender',
    'Stasik0': u'Sten Grüner',
    'Sten': u'Sten Grüner',
    'Frank Meerkoetter': u'Frank Meerkötter',
    'ichrispa': u'Chris Iatrou',
    'Chris Paul Iatrou': u'Chris Iatrou',
    'Torben-D': u'TorbenD',
    'FlorianPalm': u'Florian Palm',
    'ChristianFimmers': u'Christian Fimmers'
}

def buildFileStats(repo):

    fileRenameMap = dict()
    renamePattern = re.compile(r"(.*){(.*) => (.*)}(.*)")

    cnt = 0
    for commit in repo.iter_commits():
        cnt += 1

    curr = 0
    for commit in repo.iter_commits():
        curr += 1
        print("Checking commit {}/{}  ->   {}".format(curr, cnt, commit.hexsha))

        for objpath, stats in commit.stats.files.items():

            match = renamePattern.match(objpath)

            if match:
                # the file was renamed, store the rename to follow up later
                oldFile = (match.group(1) + match.group(2) + match.group(4)).replace("//", "/")
                newFile = (match.group(1) + match.group(3) + match.group(4)).replace("//", "/")

                while newFile in fileRenameMap:
                    newFile = fileRenameMap[newFile]

                if oldFile != newFile:
                    fileRenameMap[oldFile] = newFile
            else:
                newFile = fileRenameMap[objpath] if objpath in fileRenameMap else objpath

            if stats['insertions'] > 0:
                if not newFile in fileAuthorStats:
                    fileAuthorStats[newFile] = dict()

                authorName = unicode(commit.author.name)
                if authorName in assumeSameAuthor:
                    authorName = assumeSameAuthor[authorName]

                if not authorName in fileAuthorStats[newFile]:
                    fileAuthorStats[newFile][authorName] = {
                        'years': dict(),
                        'first_commit': commit.committed_datetime
                    }
                elif commit.committed_datetime < fileAuthorStats[newFile][authorName]['first_commit']:
                    fileAuthorStats[newFile][authorName]['first_commit'] = commit.committed_datetime

                if not commit.committed_datetime.year in fileAuthorStats[newFile][authorName]['years']:
                    fileAuthorStats[newFile][authorName]['years'][commit.committed_datetime.year] = 0

                fileAuthorStats[newFile][authorName]['years'][commit.committed_datetime.year] += stats['insertions']




def walkFiles(repo, folder, pattern):
    patternCompiled = re.compile(pattern)
    for root, subdirs, files in os.walk(folder):
        for f in files:
            if patternCompiled.match(f):
                fname = os.path.join(root,f)
                updateCopyright(repo, fname)

if __name__ == '__main__':
    baseDir = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
    repo = Repo(baseDir)
    assert not repo.bare

    buildFileStats(repo)

    dirs = ['src', 'plugins', 'include']

    for dir in dirs:
        walkFiles(repo, os.path.join(baseDir, dir), r"(.*\.c|.*\.h)$")
